[Libguestfs] [PATCH 1/5] daemon: Replace root_mounted global with intelligence.

Richard W.M. Jones rjones at redhat.com
Thu Jan 27 18:02:51 UTC 2011


This is a good general patch -- we should apply it anyway
regardless of the rest.

Rich.

-- 
Richard Jones, Virtualization Group, Red Hat http://people.redhat.com/~rjones
libguestfs lets you edit virtual machines.  Supports shell scripting,
bindings from many languages.  http://libguestfs.org
-------------- next part --------------
>From 701ba942339ab3c56b361891496cda6bd2fdbd71 Mon Sep 17 00:00:00 2001
From: Richard W.M. Jones <rjones at redhat.com>
Date: Thu, 27 Jan 2011 17:27:41 +0000
Subject: [PATCH 1/5] daemon: Replace root_mounted global with intelligence.

We used to maintain a global flag 'root_mounted' which tells us if the
user has mounted something on root (ie. on the sysroot directory).

This flag caused a lot of trouble (eg. RHBZ#599503) because it's hard
to keep the flag updated correctly when the user can do arbitrary
mounts and also use mkmountpoint.

Remove this flag and replace it with a test to see if something is
mounted on *or under* the sysroot.  (It has to be *or under* because
of mkmountpoint and friends).

This also replaces a rather convoluted "have we mounted root yet"
check in the mount* APIs with a simpler check to see if the mountpoint
exists and is an ordinary directory.
---
 daemon/daemon.h |    4 +-
 daemon/mount.c  |   74 ++++++++++++++++++++++++++++++++++++------------------
 2 files changed, 51 insertions(+), 27 deletions(-)

diff --git a/daemon/daemon.h b/daemon/daemon.h
index 6845e1b..da991b1 100644
--- a/daemon/daemon.h
+++ b/daemon/daemon.h
@@ -107,7 +107,7 @@ extern uint64_t progress_hint;
 extern uint64_t optargs_bitmask;
 
 /*-- in mount.c --*/
-extern int root_mounted;
+extern int is_root_mounted (void);
 
 /*-- in stubs.c (auto-generated) --*/
 extern void dispatch_incoming_message (XDR *);
@@ -178,7 +178,7 @@ extern void notify_progress (uint64_t position, uint64_t total);
  */
 #define NEED_ROOT(cancel_stmt,fail_stmt)                                \
   do {									\
-    if (!root_mounted) {						\
+    if (!is_root_mounted ()) {						\
       if ((cancel_stmt) != -2)                                          \
         reply_with_error ("%s: you must call 'mount' first to mount the root filesystem", __func__); \
       fail_stmt;							\
diff --git a/daemon/mount.c b/daemon/mount.c
index ccd07c6..c584f81 100644
--- a/daemon/mount.c
+++ b/daemon/mount.c
@@ -24,12 +24,43 @@
 #include <unistd.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <mntent.h>
 
 #include "daemon.h"
 #include "actions.h"
 
-/* You must mount something on "/" first, hence: */
-int root_mounted = 0;
+/* You must mount something on "/" first before many operations.
+ * Hence we have an internal function which can test if something is
+ * mounted on *or under* the sysroot directory.  (It has to be *or
+ * under* because of mkmountpoint and friends).
+ */
+int
+is_root_mounted (void)
+{
+  FILE *fp;
+  struct mntent *m;
+
+  fp = setmntent ("/etc/mtab", "r");
+  if (fp == NULL) {
+    perror ("/etc/mtab");
+    exit (EXIT_FAILURE);
+  }
+
+  while ((m = getmntent (fp)) != NULL) {
+    /* Allow a mount directory like "/sysroot". */
+    if (sysroot_len > 0 && STREQ (m->mnt_dir, sysroot)) {
+    gotit:
+      endmntent (fp);
+      return 1;
+    }
+    /* Or allow a mount directory like "/sysroot/...". */
+    if (STRPREFIX (m->mnt_dir, sysroot) && m->mnt_dir[sysroot_len] == '/')
+      goto gotit;
+  }
+
+  endmntent (fp);
+  return 0;
+}
 
 /* The "simple mount" call offers no complex options, you can just
  * mount a device on a mountpoint.  The variations like mount_ro,
@@ -44,25 +75,31 @@ int
 do_mount_vfs (const char *options, const char *vfstype,
               const char *device, const char *mountpoint)
 {
-  int r, is_root;
+  int r;
   char *mp;
   char *error;
+  struct stat statbuf;
 
   ABS_PATH (mountpoint, 0, return -1);
 
-  is_root = STREQ (mountpoint, "/");
-
-  if (!root_mounted && !is_root) {
-    reply_with_error ("you must mount something on / first");
-    return -1;
-  }
-
   mp = sysroot_path (mountpoint);
   if (!mp) {
     reply_with_perror ("malloc");
     return -1;
   }
 
+  /* Check the mountpoint exists and is a directory. */
+  if (stat (mp, &statbuf) == -1) {
+    reply_with_perror ("mount: %s", mountpoint);
+    free (mp);
+    return -1;
+  }
+  if (!S_ISDIR (statbuf.st_mode)) {
+    reply_with_perror ("mount: %s: mount point is not a directory", mountpoint);
+    free (mp);
+    return -1;
+  }
+
   if (vfstype)
     r = command (NULL, &error,
                  "mount", "-o", options, "-t", vfstype, device, mp, NULL);
@@ -76,9 +113,6 @@ do_mount_vfs (const char *options, const char *vfstype,
     return -1;
   }
 
-  if (is_root)
-    root_mounted = 1;
-
   return 0;
 }
 
@@ -134,8 +168,6 @@ do_umount (const char *pathordevice)
 
   free (err);
 
-  /* update root_mounted? */
-
   return 0;
 }
 
@@ -324,9 +356,6 @@ do_umount_all (void)
 
   free_stringslen (mounts, size);
 
-  /* We've unmounted root now, so ... */
-  root_mounted = 0;
-
   return 0;
 }
 
@@ -368,8 +397,8 @@ do_mount_loop (const char *file, const char *mountpoint)
 }
 
 /* Specialized calls mkmountpoint and rmmountpoint are really
  * variations on mkdir and rmdir which do no checking and (in the
  * mkmountpoint case) set the root_mounted flag.
  */
 int
 do_mkmountpoint (const char *path)
@@ -388,11 +417,6 @@ do_mkmountpoint (const char *path)
     return -1;
   }
 
-  /* Set the flag so that filesystems can be mounted here,
-   * not just on /sysroot.
-   */
-  root_mounted = 1;
-
   return 0;
 }
 
@@ -368,8 +397,8 @@ do_mount_loop (const char *file, const char *mountpoint)
 
 /* Specialized calls mkmountpoint and rmmountpoint are really
- * variations on mkdir and rmdir which do no checking and (in the
- * mkmountpoint case) set the root_mounted flag.
+ * variations on mkdir and rmdir which do no checking of the
+ * is_root_mounted() flag.
  */
 int
-- 
1.7.3.5



More information about the Libguestfs mailing list